#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Quotas for resources per project"""

from oslo_log import log as logging
from oslo_utils import importutils

from zun.common import exception
from zun.common import utils
import zun.conf
from zun import objects

LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


class DbQuotaDriver(object):
    """Driver to perform necessary checks to enforce quotas and obtain
    quota information. The default driver utilizes the local database.
    """

    def get_by_project(self, context, project_id, resource):
        """Get a specify quota by project."""
        return objects.Quota.get(context, project_id, resource)

    def get_by_class(self, context, quota_class, resource):
        """Get a specifiy quota by quota class."""
        return objects.QuotaClass.get(context, quota_class, resource)

    def get_defaults(self, context, resources):
        """Given a list of resource, retrieve the default quotas.
        Use the class quotas named `DEFAULT_QUOTA_CLASS_NAME` as default
        quotas if it exists.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        """
        quotas = {}
        default_quotas = objects.QuotaClass.get_all(context)
        for resource in resources.values():
            quotas[resource.name] = default_quotas.get(resource.name,
                                                       resource.default)

        return quotas

    def get_class_quotas(self, context, resources, quota_class,
                         defaults=True):
        """Given a list of resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registerd resources.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the default value will be reported
                         if there is no specific value for the resource.
        """
        quotas = {}
        class_quotas = objects.QuotaClass.get_all(context, quota_class)

        for resource in resources.values():
            if defaults or resource.name in class_quotas:
                quotas[resource.name] = class_quotas.get(resource.name,
                                                         resource.default)

        return quotas

    def _process_quotas(self, context, resources, project_id, quotas,
                        quota_class=None, defaults=True,
                        usages=None):
        modified_quotas = {}
        if quota_class:
            class_quotas = objects.QuotaClass.get_all(context,
                                                      quota_class)
        else:
            class_quotas = {}

        default_quotas = self.get_defaults(context, resources)

        for resource in resources.values():
            # Omit default quota class values
            if not defaults and resource.name not in quotas:
                continue

            limit = quotas.get(resource.name, class_quotas.get(
                resource.name, default_quotas[resource.name]))
            modified_quotas[resource.name] = dict(limit=limit)

            # Include usages if desired. This is optional because one
            # internal consumer of this interface wants to access the
            # usages directly from inside a transaction.
            if usages:
                modified_quotas[resource.name].update(
                    in_use=usages.get(resource.name, 0))

        return modified_quotas

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None, defaults=True,
                           usages=True, project_quotas=None):
        """Given a list of resources, retrieve the quotas for the given
        project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: The quota class name.
        :param defaults: If True, the quota class value (or the default value
                         ,if there is no value from the quota class) will
                         be reported if there is no specific value for the
                         resource.
        :param usages: If True, the current counts will be returned.
        :param project_quotas: Quotas dictionary for the specified project.
        """
        project_quotas = project_quotas or objects.Quota.get_all(context,
                                                                 project_id)
        project_usages = {}
        if usages:
            for resource in resources.values():
                project_usages[resource.name] = resource.count(context,
                                                               project_id)
        return self._process_quotas(context, resources, project_id,
                                    project_quotas, quota_class,
                                    defaults=defaults, usages=project_usages)

    def _get_quotas(self, context, resources, keys, project_id=None,
                    project_quotas=None):
        """A helper method which retrieves the quotas for the specific
        resources identified by keys, and which apply to the current
        context.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param keys: A list of the desired quotas to retrieve.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's project.
        :param project_quotas: Quotas dictionary for the specified project.
        """

        # Filter resources
        desired = set(keys)
        sub_resources = {k: v for k, v in resources.items() if k in desired}

        # Make sure we accounted for all of them...
        LOG.debug('Getting quotas for project %(project_id)s. Resources:'
                  '%(keys)s', {'project_id': project_id, 'keys': keys})
        # Grab and return the quotas (without usages)
        quotas = self.get_project_quotas(context, sub_resources,
                                         project_id,
                                         usages=False,
                                         project_quotas=project_quotas)

        return {k: v['limit'] for k, v in quotas.items()}

    def limit_check(self, context, resources, values, project_id=None):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown if it is not a simple limit resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high. Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registerd resources.
        :param values: A dictionary of the values to check against the
                       quota.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on common
                           user's project.
        """
        _valid_method_call_check_resources(values, 'check', resources)

        # Ensure no value is less than zero
        unders = [key for key, val in values.items() if val < 0]
        if unders:
            raise exception.InvalidQuotaValue(unders=sorted(unders))

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id

        # Get applicable quotas
        project_quotas = objects.Quota.get_all(context, project_id)
        quotas = self._get_quotas(context, resources, values.keys(),
                                  project_id=project_id,
                                  project_quotas=project_quotas)
        # Check the quotas and construct a list of the resources that
        # would be put over limit by the desired values
        overs = [key for key, val in values.items()
                 if quotas[key] >= 0 and utils.is_less_than(quotas[key], val)]

        if overs:
            headroom = {}
            for key in overs:
                headroom[key] = min(
                    val for val in (quotas.get(key), project_quotas.get(key))
                    if val is not None
                )
            raise exception.OverQuota(overs=sorted(overs), quotas=quotas,
                                      usages={}, headroom=headroom)

    def destroy_all_by_project(self, context, project_id):
        """Destroy all quotas associated with a project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        """
        objects.Quota.destroy_all_by_project(context, project_id)


class NoopQuotaDriver(object):
    """Driver that turns quotas calls into no-ops and pretends that quotas
    for all resources are unlimited. This can be used if you do not wish to
    have any quota checking.
    """

    def get_by_project(self, context, project_id, resource):
        """Get a specific quota by project."""
        # Unlimited
        return -1

    def get_by_class(self, context, quota_class, resource):
        """Get a specific quota by quota class."""
        # Unlimited
        return -1

    def get_defaults(self, context, resources):
        """Given a list of resources, retrieve the default quotas.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = -1
        return quotas

    def get_class_quotas(self, context, resources, quota_class, defaults=True):
        """Given a list resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registerd resources.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the default value will be reported
                         if there is no specific value for the
                         resource.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = -1
        return quotas

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None, defaults=None,
                           usages=True):
        """Given a list of resources, retrieve the quotas for the given
        project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the quota class value (or the default value,
                         if there is no value from the quota class) will be
                         reported if there is no specific value for
                         the resource.
        :param usage: If True, the current counts will also be returned.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = {}
            quotas[resource.name]['limit'] = -1
            if usages:
                quotas[resource.name]['in_use'] = -1
        return quotas

    def limit_check(self, context, resources, values, project_id=None):
        """Check simple quota limits.

        For limit--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of
        the resources which are too high. Otherwise, the method returns
        nothing.

        :param context: The request context, for acccess checks.
        :param resources: A dictionary of the registered resources.
        :param values: A dictionary of the values to check against
                       the quota.
        :param project_id: Specify the project_id if current context is
                           admin and the admin wants to impact on
                           common user's project.
        """
        pass

    def destroy_all_by_project(self, context, project_id):
        """Destroy all quotas associated with a project.

        :param context: The request context, for access checks.
        :param project_id: The ID the project being deleted.
        """
        pass


class BaseResource(object):
    """Describe a single resource for quota checking."""

    def __init__(self, name, flag=None):
        """Initalizes a Resource.

        :param name: The name of the resource, i.e., "containers".
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """
        self.name = name
        self.flag = flag

    def quota(self, driver, context, **kwargs):
        """Given a driver and context, obtains the quota for this
        resource.

        :param driver: A quota driver.
        :param context: The request context.
        :param project_id: The project to obtain the quota value for.
                           If not provided, it is taken from the context.
                           If it is given as None, no project-specific
                           quota will be searched for.
        :param quota_class: The name of the quota class to return
                            quotas for.
        """
        # Get the project ID
        project_id = kwargs.get('project_id', context.project_id)

        # Get the quota class
        quota_class = kwargs.get('quota_class')

        # Look up the quota for the project
        if project_id:
            try:
                return driver.get_by_project(context, project_id, self.name)
            except exception.ProjectQuotaNotFound:
                pass

        # Try for the quota class
        if quota_class:
            try:
                return driver.get_by_class(context, quota_class, self.name)
            except exception.QuotaClassNotFound:
                pass

        return self.default

    @property
    def default(self):
        """Return the default value of the quota."""
        return CONF.quota[self.flag] if self.flag else -1


class AbsoluteResource(BaseResource):
    """Describe a resource that does not correspond to database objects."""
    pass


class CountableResource(BaseResource):
    """Describe a resource where the counts aren't based solely on the
    project ID.
    """

    def __init__(self, name, count, flag=None):
        """Initalizes a CountableResource.

        Countable resources are those resources which directly
        correspond to objects in the database, but for which a count
        by project ID is inappropriate.

        A CountableResource must be constructed with a counting
        function, which will be called to determine the current counts
        of the resource.

        The counting function will be passed the context, along with
        the extra positional and keyword arguments that are passed to
        Quota.count_(). It should return an integer specifying the
        count scoped to a project.

        Note that this counting is not performed a transaction-safe
        manner.

        :param name: The name of the resource, i.e., "containers"
        :param count: A callable which returns the count of the
                      resource. The arguments passed are as described
                      above.
        :param flag: The name of the flag or configuration option
                     which specfies the default value of the quota
                     for this resource.
        """
        super(CountableResource, self).__init__(name, flag=flag)
        self._count = count

    def count(self, context, project_id):
        return self._count(context, project_id)


class QuotaEngine(object):
    """Represent the set of recognized quotas."""

    def __init__(self, quota_driver=None):
        """Initalize a Quota object."""
        self._resources = {}
        self.__driver = quota_driver

    @property
    def _driver(self):
        if self.__driver:
            return self.__driver

        self.__driver = importutils.import_object(CONF.quota.driver)
        return self.__driver

    def register_resource(self, resource):
        """Register a resource"""

        self._resources[resource.name] = resource

    def register_resources(self, resources):
        """Register a list of resources."""

        for resource in resources:
            self.register_resource(resource)

    def get_by_project(self, context, project_id, resource):
        """Get a specific quota by project."""

        return self._driver.get_by_project(context, project_id, resource)

    def get_by_class(self, context, quota_class, resource):
        """Get a specific quota by quota class."""

        return self._driver.get_by_class(context, quota_class, resource)

    def get_defaults(self, context):
        """Retrieve the default quotas.

        :param context: The request context, for access checks.
        """

        return self._driver.get_defaults(context, self._resources)

    def get_class_quotas(self, context, quota_class, defaults=True):
        """Retrieve the quotas for the given quota class.

        :param context: The request context, for access checks.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the default value will be reported
                         if there is no specific value for the resource.
        """

        return self._driver.get_class_quotas(context, self._resources,
                                             quota_class, defaults=defaults)

    def get_project_quotas(self, context, project_id, quota_class=None,
                           defaults=True, usages=True):
        """Retrieve the quotas for the given project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the quota class value (or the default
                         value, if there is no value from the quota class)
                         will be reported if there is no specific value
                         for the resource.
        :param usages: If True, the current counts will also be returned.
        """

        return self._driver.get_project_quotas(context, self._resources,
                                               project_id,
                                               quota_class=quota_class,
                                               defaults=defaults,
                                               usages=usages)

    def count(self, context, resource, project_id):
        """Count a resource.

        For countable resources, invokes the count() function and
        returns its result. Argument following the context and
        resource are passed directly to the count function declared
        by the resource.

        :param context: The request context, for access check.
        :param resource: The name of the resource, as a string.
        """

        # Get the resource
        res = self._resources.get(resource)
        if not res or not hasattr(res, 'count'):
            raise exception.QuotaResourceUnknown(unknown=[resource])

        return res.count(context, project_id)

    def limit_check(self, context, project_id=None, **values):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction. The
        values to check are given as keyword arguments, where the key
        identifies the specific quota limit to check, and the value
        is the proposed value.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit resource.

        If any of proposed values is over the defined quota, an OverQuota
        exception will be raised with the sorted list of the resources
        which are too high. Otherwise, the method returns nothing.

        :param context: The rquest context, for access checks.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's project.
        """

        return self._driver.limit_check(context, self._resources, values,
                                        project_id=project_id)

    def destroy_all_by_project(self, context, project_id):
        """Destroy all quotas, usages associated with a project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        """

        self._driver.destroy_all_by_project(context, project_id)

    @property
    def resources(self):
        if isinstance(self._driver, NoopQuotaDriver):
            return -1
        return sorted(self._resources.keys())


def _containers_count(context, project_id):
    return objects.Container.get_count(context, project_id, flag='containers')


def _cpu_count(context, project_id):
    return round(objects.Container.get_count(
        context, project_id, flag='cpu'), 3)


def _disk_count(context, project_id):
    return objects.Container.get_count(context, project_id, flag='disk')


def _memory_count(context, project_id):
    return objects.Container.get_count(context, project_id, flag='memory')


QUOTAS = QuotaEngine()


resources = [
    CountableResource('containers', _containers_count, flag='containers'),
    CountableResource('memory', _memory_count, flag='memory'),
    CountableResource('disk', _disk_count, flag='disk'),
    CountableResource('cpu', _cpu_count, flag='cpu')
]


QUOTAS.register_resources(resources)


def _valid_method_call_check_resource(name, method, resources):
    if name not in resources:
        raise exception.InvalidQuotaMethodUsage(method=method, res=name)


def _valid_method_call_check_resources(resource_values, method, resources):
    """A method to check whether the resource can use the quota method.

    :param resource_values: Dict containing the resource names and values.
    :param method: The quota method to check.
    :param resources: Dict containing Resource objects to validate against.
    """

    for name in resource_values.keys():
        _valid_method_call_check_resource(name, method, resources)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Convert type of 'command' from string to list

Revision ID: 3e80bbfd8da7
Revises: 26896d5f9053
Create Date: 2018-06-20 11:21:38.077673

"""

import shlex

from alembic import op
import sqlalchemy as sa

from oslo_serialization import jsonutils


# revision identifiers, used by Alembic.
revision = '3e80bbfd8da7'
down_revision = '26896d5f9053'
branch_labels = None
depends_on = None


TABLE_MODEL = sa.Table(
    'container', sa.MetaData(),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('command', sa.Text()))


def upgrade():
    op.alter_column('container', 'command', type_=sa.Text())
    # Convert 'command' from string to json-encoded list
    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        for row in session.query(TABLE_MODEL):
            if row[1]:
                command = shlex.split(row[1])
                command = jsonutils.dumps(command)
                session.execute(
                    TABLE_MODEL.update().values(
                        command=command).where(
                            TABLE_MODEL.c.id == row[0]))
    session.commit()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

availability_zone_opts = [
    cfg.StrOpt('default_availability_zone',
               default='nova',
               help="""
Default availability zone for compute services.

This option determines the default availability zone for 'zun-compute'
services.

Possible values:

* Any string representing an existing availability zone name.
"""),
    cfg.StrOpt('default_schedule_zone',
               help="""
Default availability zone for containers.

This option determines the default availability zone for containers, which will
be used when a user does not specify one when creating a container. The
container(s) will be bound to this availability zone for their lifetime.

Possible values:

* Any string representing an existing availability zone name.
* None, which means that the container can move from one availability zone to
  another during its lifetime if it is moved from one compute node to another.
"""),
]


def register_opts(conf):
    conf.register_opts(availability_zone_opts)


def list_opts():
    return {'DEFAULT': availability_zone_opts}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import zun.conf
from zun.db import api as db_api
from zun.db.sqlalchemy import api as sqla_api
from zun.db.sqlalchemy import migration
from zun.tests import base
from zun.tests.unit.db.base import Database

CONF = zun.conf.CONF

_DB_CACHE = None


class DriverTestCase(base.TestCase):
    def setUp(self):
        super(DriverTestCase, self).setUp()
        self.dbapi = db_api._get_dbdriver_instance()

        global _DB_CACHE
        if not _DB_CACHE:
            _DB_CACHE = Database(sqla_api, migration,
                                 sql_connection=CONF.database.connection)
        self.useFixture(_DB_CACHE)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""support cpuset

Revision ID: 2b129060baff
Revises: 33cdd98bb9b2
Create Date: 2018-011-10 10:08:40.547664

"""

# revision identifiers, used by Alembic.
revision = '2b129060baff'
down_revision = '33cdd98bb9b2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from zun.db.sqlalchemy import models


def upgrade():
    op.add_column('container',
                  sa.Column('cpu_policy', sa.String(length=255)))
    op.add_column('container',
                  sa.Column('cpuset', models.JSONEncodedDict, nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add registry_id to container

Revision ID: 1bc34e18180b
Revises: 5ffc1cabe6b4
Create Date: 2019-01-06 21:45:57.505152

"""

# revision identifiers, used by Alembic.
revision = '1bc34e18180b'
down_revision = '5ffc1cabe6b4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('container',
                  sa.Column('registry_id', sa.Integer(),
                            nullable=True))
    op.create_foreign_key(
        None, 'container', 'registry', ['registry_id'], ['id'])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

docker_group = cfg.OptGroup(name='docker',
                            title='Options for docker')

docker_opts = [
    cfg.StrOpt('docker_remote_api_version',
               default='1.26',
               help='Docker remote api version. Override it according to '
                    'specific docker api version in your environment.'),
    cfg.IntOpt('default_timeout',
               default=60,
               help='Default timeout in seconds for docker client '
                    'operations.'),
    cfg.StrOpt('api_url',
               default='unix:///var/run/docker.sock',
               help='API endpoint of docker daemon'),
    cfg.StrOpt('docker_remote_api_url',
               default='tcp://$docker_remote_api_host:$docker_remote_api_port',
               help='Remote API endpoint of docker daemon'),
    cfg.BoolOpt('api_insecure',
                default=False,
                help='If set, ignore any SSL validation issues'),
    cfg.StrOpt('ca_file',
               help='Location of CA certificates file for '
                    'securing docker api requests (tlscacert).'),
    cfg.StrOpt('cert_file',
               help='Location of TLS certificate file for '
                    'securing docker api requests (tlscert).'),
    cfg.StrOpt('key_file',
               help='Location of TLS private key file for '
                    'securing docker api requests (tlskey).'),
    cfg.StrOpt('docker_remote_api_host',
               default='$my_ip',
               help='Defines the remote api host for the docker daemon.'),
    cfg.StrOpt('docker_remote_api_port',
               default='2375',
               help='Defines the remote api port for the docker daemon.'),
    cfg.IntOpt('execute_timeout',
               default=5,
               help='Timeout in seconds for executing a command in a docker '
                    'container.'),
    cfg.StrOpt('docker_data_root',
               default='/var/lib/docker',
               deprecated_for_removal=True,
               help='Root directory of persistent Docker state.'),
    cfg.StrOpt('default_registry',
               help='The default registry from which docker images are '
                    'pulled. Its value can be the registry domain name '
                    '(e.g. docker.io) or None.'),
    cfg.StrOpt('default_registry_username',
               help='The username of the default registry.'),
    cfg.StrOpt('default_registry_password',
               help='The password of the default registry.'),
]

ALL_OPTS = (docker_opts)


def register_opts(conf):
    conf.register_group(docker_group)
    conf.register_opts(ALL_OPTS, docker_group)


def list_opts():
    return {docker_group: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_log import log as logging
from oslo_utils import excutils

from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
import zun.conf


PROC_MOUNTS_PATH = '/proc/mounts'

LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


class MountInfo(object):
    def __init__(self, device, mountpoint, fstype, opts):
        self.device = device
        self.mountpoint = mountpoint
        self.fstype = fstype
        self.opts = opts

    def __repr__(self, *args, **kwargs):
        return str(self.__dict__)


class Mounter(object):
    def make_filesystem(self, devpath, fstype):
        try:
            utils.execute('mkfs', '-t', fstype, '-F', devpath,
                          run_as_root=True)
        except exception.CommandError as e:
            raise exception.MakeFileSystemException(_(
                "Unexpected error while make filesystem. "
                "Devpath: %(devpath)s, "
                "Fstype: %(fstype)s, "
                "Error: %(error)s") %
                {'devpath': devpath, 'fstype': fstype, 'error': e})

    def mount(self, devpath, mountpoint, fstype=None):
        try:
            utils.execute('mount', '-t', fstype, devpath, mountpoint,
                          run_as_root=True)
        except exception.CommandError as e:
            raise exception.MountException(_(
                "Unexpected error while mount block device. "
                "Devpath: %(devpath)s, "
                "Mountpoint: %(mountpoint)s, "
                "Error: %(error)s") %
                {'devpath': devpath, 'mountpoint': mountpoint, 'error': e})

    def unmount(self, mountpoint):
        try:
            utils.execute('umount', mountpoint, run_as_root=True)
        except exception.CommandError as e:
            raise exception.UnmountException(_(
                "Unexpected err while unmount block device. "
                "Mountpoint: %(mountpoint)s, "
                "Error: %(error)s") %
                {'mountpoint': mountpoint, 'error': e})

    def read_mounts(self, filter_device=None, filter_fstype=None):
        """Read all mounted filesystems.

        Read all mounted filesystems except filtered option.

        :param filter_device: Filter for device, the result will not contain
                              the mounts whose device argument in it.
        :param filter_fstype: Filter for mount point.
        :return: All mounts.
        """
        if filter_device is None:
            filter_device = ()
        if filter_fstype is None:
            filter_fstype = ()

        try:
            (out, err) = utils.execute('cat', PROC_MOUNTS_PATH,
                                       check_exit_code=0)
        except exception.CommandError:
            msg = _("Failed to read mounts.")
            raise exception.FileNotFound(msg)

        lines = out.split('\n')
        mounts = []
        for line in lines:
            if not line:
                continue
            tokens = line.split()
            if len(tokens) < 4:
                continue
            if tokens[0] in filter_device or tokens[1] in filter_fstype:
                continue
            mounts.append(MountInfo(device=tokens[0], mountpoint=tokens[1],
                                    fstype=tokens[2], opts=tokens[3]))
        return mounts

    def get_mps_by_device(self, devpath):
        """Get all mountpoints that device mounted on.

        :param devpath: The path of mount device.
        :return: All mountpoints.
        """
        mps = []
        mounts = self.read_mounts()
        for m in mounts:
            if devpath == m.device:
                mps.append(m.mountpoint)
        return mps


def check_already_mounted(mountpoint):
    """Check that the mount device is mounted on the specific mount point.

    :param devpath: The path of mount deivce.
    :param mountpoint: The path of mount point.
    :rtype: bool
    """
    mounts = Mounter().read_mounts()
    for m in mounts:
        if mountpoint == m.mountpoint:
            return True
    return False


def do_mount(devpath, mountpoint, fstype):
    """Execute device mount operation.

    :param devpath: The path of mount device.
    :param mountpoint: The path of mount point.
    :param fstype: The file system type.
    """
    if check_already_mounted(mountpoint):
        return

    mounter = Mounter()
    try:
        mounter.mount(devpath, mountpoint, fstype)
    except exception.MountException:
        try:
            mounter.make_filesystem(devpath, fstype)
            mounter.mount(devpath, mountpoint, fstype)
        except exception.ZunException as e:
            with excutils.save_and_reraise_exception():
                LOG.error(e.message)


def do_unmount(mountpoint):
    if not check_already_mounted(mountpoint):
        return
    Mounter().unmount(mountpoint)


def get_mountpoint(volume_id):
    return os.path.join(CONF.volume.volume_dir, volume_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from zun.db import api as dbapi
from zun.objects import base
from zun.objects import fields as z_fields


@base.ZunObjectRegistry.register
class ResourceClass(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    # Version 1.1: Add uuid field
    VERSION = '1.1'

    fields = {
        'id': fields.IntegerField(read_only=True),
        'uuid': fields.UUIDField(nullable=False),
        'name': z_fields.ResourceClassField(nullable=False),
    }

    @staticmethod
    def _from_db_object(resource, db_resource):
        """Converts a database entity to a formal object."""
        for field in resource.fields:
            setattr(resource, field, db_resource[field])

        resource.obj_reset_changes()
        return resource

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [ResourceClass._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a resource class based on uuid.

        :param uuid: the uuid of a resource class.
        :param context: Security context
        :returns: a :class:`ResourceClass` object.
        """
        db_resource = dbapi.get_resource_class(context, uuid)
        resource = ResourceClass._from_db_object(cls(context), db_resource)
        return resource

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        """Find a resource class based on name.

        :param name: the name of a resource class.
        :param context: Security context
        :returns: a :class:`ResourceClass` object.
        """
        db_resource = dbapi.get_resource_class(context, name)
        resource = ResourceClass._from_db_object(cls(context), db_resource)
        return resource

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None, sort_key=None,
             sort_dir=None):
        """Return a list of ResourceClass objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :returns: a list of :class:`ResourceClass` object.

        """
        db_resources = dbapi.list_resource_classes(
            context, limit=limit, marker=marker, sort_key=sort_key,
            sort_dir=sort_dir)
        return ResourceClass._from_db_object_list(
            db_resources, cls, context)

    @base.remotable
    def create(self, context):
        """Create a ResourceClass record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ResourceClass(context)

        """
        values = self.obj_get_changes()
        db_resource = dbapi.create_resource_class(context, values)
        self._from_db_object(self, db_resource)

    @base.remotable
    def destroy(self, context=None):
        """Delete the ResourceClass from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ResourceClass(context)
        """
        dbapi.destroy_resource_class(context, self.uuid)
        self.obj_reset_changes()

    @base.remotable
    def save(self, context=None):
        """Save updates to this ResourceClass.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ResourceClass(context)
        """
        updates = self.obj_get_changes()
        dbapi.update_resource_class(context, self.uuid, updates)

        self.obj_reset_changes()

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this ResourceClass.

        Loads a resource class with the same id from the database and
        checks for updated attributes. Updates are applied from
        the loaded resource class column by column, if there are any
        updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ResourceClass(context)
        """
        current = self.__class__.get_by_uuid(self._context, self.uuid)
        for field in self.fields:
            if self.obj_attr_is_set(field) and \
               getattr(self, field) != getattr(current, field):
                setattr(self, field, getattr(current, field))



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add pci stats to compute node

Revision ID: ff7b9665d504
Revises: fc27c7415d9c
Create Date: 2017-09-26 13:49:11.470002

"""

# revision identifiers, used by Alembic.
revision = 'ff7b9665d504'
down_revision = 'fc27c7415d9c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('compute_node',
                  sa.Column('pci_stats', sa.Text()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from http import client as httplib
import requests
import traceback

from os_vif.objects import base
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from zun.common import consts
from zun.common import exception
from zun.common.i18n import _


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class CNIRunner(object, metaclass=abc.ABCMeta):
    # TODO(ivc): extend SUPPORTED_VERSIONS and format output based on
    # requested params.CNI_VERSION and/or params.config.cniVersion
    VERSION = '0.3.1'
    SUPPORTED_VERSIONS = ['0.3.1']

    @abc.abstractmethod
    def _add(self, params):
        raise NotImplementedError()

    @abc.abstractmethod
    def _delete(self, params):
        raise NotImplementedError()

    def _write_dict(self, fout, dct):
        output = {'cniVersion': self.VERSION}
        output.update(dct)
        LOG.debug("CNI output: %s", output)
        jsonutils.dump(output, fout, sort_keys=True)

    def _write_exception(self, fout, msg):
        self._write_dict(fout, {
            'msg': msg,
            'code': consts.CNI_EXCEPTION_CODE,
            'details': traceback.format_exc(),
        })

    def _write_version(self, fout):
        self._write_dict(fout, {'supportedVersions': self.SUPPORTED_VERSIONS})

    @abc.abstractmethod
    def prepare_env(self, env, stdin):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_container_id(self, params):
        raise NotImplementedError()

    def run(self, env, fin, fout):
        try:
            # Prepare params according to calling Object
            params = self.prepare_env(env, fin)
            if env.get('CNI_COMMAND') == 'ADD':
                vif = self._add(params)
                self._write_dict(fout, vif)
            elif env.get('CNI_COMMAND') == 'DEL':
                self._delete(params)
            elif env.get('CNI_COMMAND') == 'VERSION':
                self._write_version(fout)
            else:
                raise exception.CNIError(
                    _("unknown CNI_COMMAND: %s"), env['CNI_COMMAND'])
            return 0
        except Exception as ex:
            # LOG.exception
            self._write_exception(fout, str(ex))
            return 1

    def _vif_data(self, vif, params):
        result = {}
        nameservers = []

        cni_ip_list = result.setdefault("ips", [])
        cni_routes_list = result.setdefault("routes", [])
        result["interfaces"] = [
            {
                "name": params["CNI_IFNAME"],
                "mac": vif.address,
                "sandbox": self.get_container_id(params)}]
        for subnet in vif.network.subnets.objects:
            cni_ip = {}
            nameservers.extend(subnet.dns)

            ip = subnet.ips.objects[0].address

            cni_ip['version'] = str(ip.version)
            cni_ip['address'] = "%s/%s" % (ip, subnet.cidr.prefixlen)
            cni_ip['interface'] = len(result["interfaces"]) - 1

            if subnet.obj_attr_is_set('gateway'):
                cni_ip['gateway'] = str(subnet.gateway)

            if subnet.routes.objects:
                routes = [
                    {'dst': str(route.cidr), 'gw': str(route.gateway)}
                    for route in subnet.routes.objects]
                cni_routes_list.extend(routes)
            cni_ip_list.append(cni_ip)

        if nameservers:
            result['dns'] = {'nameservers': nameservers}
        return result


class CNIDaemonizedRunner(CNIRunner):

    def _add(self, params):
        resp = self._make_request('addNetwork', params, httplib.ACCEPTED)
        vif = base.VersionedObject.obj_from_primitive(resp.json())
        return self._vif_data(vif, params)

    def _delete(self, params):
        self._make_request('delNetwork', params, httplib.NO_CONTENT)

    def prepare_env(self, env, stdin):
        cni_envs = {}
        cni_envs.update(
            {k: v for k, v in env.items() if k.startswith('CNI_')})
        cni_envs['config_zun'] = dict(stdin)
        return cni_envs

    def get_container_id(self, params):
        return params["CNI_CONTAINERID"]

    def _make_request(self, path, cni_envs, expected_status=None):
        method = 'POST'

        host = CONF.cni_daemon.cni_daemon_host
        port = CONF.cni_daemon.cni_daemon_port
        url = 'http://%s:%s/%s' % (host, port, path)
        try:
            LOG.debug('Making request to CNI Daemon. %(method)s %(path)s\n'
                      '%(body)s',
                      {'method': method, 'path': url, 'body': cni_envs})
            resp = requests.post(url, json=cni_envs,
                                 headers={'Connection': 'close'})
        except requests.ConnectionError:
            LOG.exception('Looks like %s:%s cannot be reached. '
                          'Is zun-cni-daemon running?', (host, port))
            raise
        LOG.debug('CNI Daemon returned "%(status)d %(reason)s".',
                  {'status': resp.status_code, 'reason': resp.reason})
        if expected_status and resp.status_code != expected_status:
            LOG.error('CNI daemon returned error "%(status)d %(reason)s".',
                      {'status': resp.status_code, 'reason': resp.reason})
            raise exception.CNIError('Got invalid status code from CNI daemon')
        return resp



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import fixture as utils_fixture
from oslo_utils import timeutils

from testtools.matchers import HasLength

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

NOW = timeutils.utcnow().replace(microsecond=0)


class TestContainerActionObject(base.DbTestCase):

    def setUp(self):
        super(TestContainerActionObject, self).setUp()
        self.fake_action = utils.get_test_action()

    def test_get_by_request_id(self):
        container_ident = self.fake_action['container_uuid']
        request_id = self.fake_action['request_id']
        with mock.patch.object(self.dbapi, 'action_get_by_request_id',
                               autospec=True) as mock_get_action:
            mock_get_action.return_value = self.fake_action
            action = objects.ContainerAction.get_by_request_id(
                self.context, container_ident, request_id)
            mock_get_action.assert_called_once_with(
                self.context, container_ident, request_id)
            self.assertEqual(self.context, action._context)

    def test_get_by_container_uuid(self):
        container_ident = self.fake_action['container_uuid']
        with mock.patch.object(self.dbapi, 'actions_get', autospec=True) \
            as mock_get_actions:
            mock_get_actions.return_value = [self.fake_action]
            actions = objects.ContainerAction.get_by_container_uuid(
                self.context, container_ident)
            mock_get_actions.assert_called_once_with(self.context,
                                                     container_ident)

            self.assertThat(actions, HasLength(1))
            self.assertIsInstance(actions[0], objects.ContainerAction)
            self.assertEqual(self.context, actions[0]._context)

    def test_action_start(self):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        container_ident = self.fake_action['container_uuid']
        action_name = self.fake_action['action']
        test_class = objects.ContainerAction
        expected_packed_values = test_class.pack_action_start(
            self.context, container_ident, action_name)
        with mock.patch.object(self.dbapi, 'action_start', autospec=True) \
            as mock_action_start:
            mock_action_start.return_value = self.fake_action
            action = objects.ContainerAction.action_start(
                self.context, container_ident, action_name, want_result=True)
            mock_action_start.assert_called_once_with(
                self.context, expected_packed_values)
            self.assertEqual(self.context, action._context)


class TestContainerActionEventObject(base.DbTestCase):

    def setUp(self):
        super(TestContainerActionEventObject, self).setUp()
        self.fake_action = utils.get_test_action()
        self.fake_event = utils.get_test_action_event()

    def test_get_by_action(self):
        action_id = self.fake_event['action_id']
        with mock.patch.object(self.dbapi, 'action_events_get',
                               autospec=True) as mock_get_event:
            mock_get_event.return_value = [self.fake_event]
            events = objects.ContainerActionEvent.get_by_action(self.context,
                                                                action_id)
            mock_get_event.assert_called_once_with(self.context, action_id)
            self.assertThat(events, HasLength(1))
            self.assertIsInstance(events[0], objects.ContainerActionEvent)
            self.assertEqual(self.context, events[0]._context)

    def test_event_start(self):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        container_uuid = self.fake_action['container_uuid']
        event_name = self.fake_event['event']
        test_class = objects.ContainerActionEvent
        expected_packed_values = test_class.pack_action_event_start(
            self.context, container_uuid, event_name)
        with mock.patch.object(self.dbapi, 'action_event_start',
                               autospec=True) as mock_event_start:
            mock_event_start.return_value = self.fake_event
            event = objects.ContainerActionEvent.event_start(
                self.context, container_uuid, event_name, want_result=True)
            mock_event_start.assert_called_once_with(self.context,
                                                     expected_packed_values)
            self.assertEqual(self.context, event._context)

    def test_event_finish(self):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        container_uuid = self.fake_action['container_uuid']
        event_name = self.fake_event['event']
        test_class = objects.ContainerActionEvent
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, container_uuid, event_name)
        with mock.patch.object(self.dbapi, 'action_event_finish',
                               autospec=True) as mock_event_finish:
            mock_event_finish.return_value = self.fake_event
            event = objects.ContainerActionEvent.event_finish(
                self.context, container_uuid, event_name, want_result=True)
            mock_event_finish.assert_called_once_with(self.context,
                                                      expected_packed_values)
            self.assertEqual(self.context, event._context)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
from oslo_log import log as logging
import tenacity

from zun.common import consts
from zun.common import context as zun_context
from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
import zun.conf
from zun.container import driver
from zun.criapi import api_pb2
from zun.criapi import api_pb2_grpc
from zun.network import neutron
from zun.network import os_vif_util
from zun import objects


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


class CriDriver(driver.BaseDriver, driver.CapsuleDriver):
    """Implementation of container drivers for CRI runtime."""

    # TODO(hongbin): define a list of capabilities of this driver.
    capabilities = {}

    def __init__(self):
        super(CriDriver, self).__init__()
        channel = grpc.insecure_channel(
            'unix:///run/containerd/containerd.sock')
        self.runtime_stub = api_pb2_grpc.RuntimeServiceStub(channel)
        self.image_stub = api_pb2_grpc.ImageServiceStub(channel)

    def create_capsule(self, context, capsule, image, requested_networks,
                       requested_volumes):

        self._create_pod_sandbox(context, capsule, requested_networks)

        # TODO(hongbin): handle init containers
        for container in capsule.init_containers:
            self._create_container(context, capsule, container,
                                   requested_networks,
                                   requested_volumes)
            self._wait_for_init_container(context, container)
            container.save(context)

        for container in capsule.containers:
            self._create_container(context, capsule, container,
                                   requested_networks,
                                   requested_volumes)
            container.status = consts.RUNNING
            container.save(context)

        capsule.status = consts.RUNNING
        return capsule

    def _create_pod_sandbox(self, context, capsule, requested_networks):
        sandbox_config = self._get_sandbox_config(capsule)
        runtime = capsule.runtime or CONF.container_runtime
        if runtime == "runc":
            # pass "" to specify the default runtime which is runc
            runtime = ""

        self._write_cni_metadata(context, capsule, requested_networks)
        sandbox_resp = self.runtime_stub.RunPodSandbox(
            api_pb2.RunPodSandboxRequest(
                config=sandbox_config,
                runtime_handler=runtime,
            )
        )
        LOG.debug("podsandbox is created: %s", sandbox_resp)
        capsule.container_id = sandbox_resp.pod_sandbox_id

    def _get_sandbox_config(self, capsule):
        return api_pb2.PodSandboxConfig(
            metadata=api_pb2.PodSandboxMetadata(
                name=capsule.uuid, namespace="default", uid=capsule.uuid
            )
        )

    def _write_cni_metadata(self, context, capsule, requested_networks):
        neutron_api = neutron.NeutronAPI(context)
        security_group_ids = utils.get_security_group_ids(
            context, capsule.security_groups)
        # TODO(hongbin): handle multiple nics
        requested_network = requested_networks[0]
        network_id = requested_network['network']
        addresses, port = neutron_api.create_or_update_port(
            capsule, network_id, requested_network, consts.DEVICE_OWNER_ZUN,
            security_group_ids, set_binding_host=True)
        capsule.addresses = {network_id: addresses}

        neutron_api = neutron.NeutronAPI(zun_context.get_admin_context())
        network = neutron_api.show_network(port['network_id'])['network']
        subnets = {}
        for fixed_ip in port['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            subnets[subnet_id] = neutron_api.show_subnet(subnet_id)['subnet']
        vif_plugin = port.get('binding:vif_type')
        vif_obj = os_vif_util.neutron_to_osvif_vif(vif_plugin, port, network,
                                                   subnets)
        state = objects.vif.VIFState(default_vif=vif_obj)
        state_dict = state.obj_to_primitive()
        capsule.cni_metadata = {consts.CNI_METADATA_VIF: state_dict}
        capsule.save(context)

    def _create_container(self, context, capsule, container,
                          requested_networks, requested_volumes):
        # pull image
        self._pull_image(context, container)

        sandbox_config = self._get_sandbox_config(capsule)
        container_config = self._get_container_config(context, container,
                                                      requested_volumes)
        response = self.runtime_stub.CreateContainer(
            api_pb2.CreateContainerRequest(
                pod_sandbox_id=capsule.container_id,
                config=container_config,
                sandbox_config=sandbox_config,
            )
        )

        LOG.debug("container is created: %s", response)
        container.container_id = response.container_id
        container.save(context)

        response = self.runtime_stub.StartContainer(
            api_pb2.StartContainerRequest(
                container_id=container.container_id
            )
        )
        LOG.debug("container is started: %s", response)

    def _get_container_config(self, context, container, requested_volumes):
        args = []
        if container.command:
            args = [str(c) for c in container.command]
        envs = []
        if container.environment:
            envs = [api_pb2.KeyValue(key=str(k), value=str(v))
                    for k, v in container.environment.items()]
        mounts = []
        if container.uuid in requested_volumes:
            req_volume = requested_volumes[container.uuid]
            mounts = self._get_mounts(context, req_volume)
        working_dir = container.workdir or ""
        labels = container.labels or []

        cpu = 0
        if container.cpu is not None:
            cpu = int(1024 * container.cpu)
        memory = 0
        if container.memory is not None:
            memory = int(container.memory) * 1024 * 1024
        linux_config = api_pb2.LinuxContainerConfig(
            security_context=api_pb2.LinuxContainerSecurityContext(
                privileged=container.privileged
            ),
            resources={
                'cpu_shares': cpu,
                'memory_limit_in_bytes': memory,
            }
        )

        # TODO(hongbin): add support for entrypoint
        return api_pb2.ContainerConfig(
            metadata=api_pb2.ContainerMetadata(name=container.name),
            image=api_pb2.ImageSpec(image=container.image),
            tty=container.tty,
            stdin=container.interactive,
            args=args,
            envs=envs,
            working_dir=working_dir,
            labels=labels,
            mounts=mounts,
            linux=linux_config,
        )

    def _pull_image(self, context, container):
        # TODO(hongbin): add support for private registry
        response = self.image_stub.PullImage(
            api_pb2.PullImageRequest(
                image=api_pb2.ImageSpec(image=container.image)
            )
        )
        LOG.debug("image is pulled: %s", response)

    def _get_mounts(self, context, volmaps):
        mounts = []
        for volume in volmaps:
            volume_driver = self._get_volume_driver(volume)
            source, destination = volume_driver.bind_mount(context, volume)
            mounts.append(api_pb2.Mount(container_path=destination,
                                        host_path=source))
        return mounts

    def _wait_for_init_container(self, context, container, timeout=3600):
        def retry_if_result_is_false(result):
            return result is False

        def check_init_container_stopped():
            status = self._show_container(context, container).status
            if status == consts.STOPPED:
                return True
            elif status == consts.RUNNING:
                return False
            else:
                raise exception.ZunException(
                    _("Container has unexpected status: %s") % status)

        r = tenacity.Retrying(
            stop=tenacity.stop_after_delay(timeout),
            wait=tenacity.wait_exponential(),
            retry=tenacity.retry_if_result(retry_if_result_is_false))
        r.call(check_init_container_stopped)

    def _show_container(self, context, container):
        container_id = container.container_id
        if not container_id:
            return

        response = self.runtime_stub.ListContainers(
            api_pb2.ListContainersRequest(
                filter={'id': container_id}
            )
        )
        if not response.containers:
            raise exception.ZunException(
                "Container %s is not found in runtime" % container_id)

        container_response = response.containers[0]
        self._populate_container(container, container_response)
        return container

    def _populate_container(self, container, response):
        self._populate_container_state(container, response)

    def _populate_container_state(self, container, response):
        state = response.state
        if state == api_pb2.ContainerState.CONTAINER_CREATED:
            container.status = consts.CREATED
        elif state == api_pb2.ContainerState.CONTAINER_RUNNING:
            container.status = consts.RUNNING
        elif state == api_pb2.ContainerState.CONTAINER_EXITED:
            container.status = consts.STOPPED
        elif state == api_pb2.ContainerState.CONTAINER_UNKNOWN:
            LOG.debug('State is unknown, status: %s', state)
            container.status = consts.UNKNOWN
        else:
            LOG.warning('Receive unexpected state from CRI runtime: %s', state)
            container.status = consts.UNKNOWN
            container.status_reason = "container state unknown"

    def delete_capsule(self, context, capsule, force):
        pod_id = capsule.container_id
        if not pod_id:
            return

        try:
            response = self.runtime_stub.StopPodSandbox(
                api_pb2.StopPodSandboxRequest(
                    pod_sandbox_id=capsule.container_id,
                )
            )
            LOG.debug("podsandbox is stopped: %s", response)
            response = self.runtime_stub.RemovePodSandbox(
                api_pb2.RemovePodSandboxRequest(
                    pod_sandbox_id=capsule.container_id,
                )
            )
            LOG.debug("podsandbox is removed: %s", response)
        except exception.CommandError as e:
            if 'error occurred when try to find sandbox' in str(e):
                LOG.error("cannot find pod sandbox in runtime")
                pass
            else:
                raise

        self._delete_neutron_ports(context, capsule)

    def _delete_neutron_ports(self, context, capsule):
        if not capsule.addresses:
            return

        neutron_ports = set()
        all_ports = set()
        for net_uuid, addrs_list in capsule.addresses.items():
            for addr in addrs_list:
                all_ports.add(addr['port'])
                if not addr['preserve_on_delete']:
                    port_id = addr['port']
                    neutron_ports.add(port_id)

        neutron_api = neutron.NeutronAPI(context)
        neutron_api.delete_or_unbind_ports(all_ports, neutron_ports)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utils for testing the API service.
"""
import datetime

import pytz


def zservice_get_data(**kwargs):
    """Simulate what the RPC layer will get from DB """
    faketime = datetime.datetime(2001, 1, 1, tzinfo=pytz.UTC)
    return {
        'binary': kwargs.get('binary', 'fake-binary'),
        'host': kwargs.get('host', 'fake-host'),
        'id': kwargs.get('id', 13),
        'report_count': kwargs.get('report_count', 13),
        'disabled': kwargs.get('disabled', False),
        'disabled_reason': kwargs.get('disabled_reason'),
        'forced_down': kwargs.get('forced_down', False),
        'last_seen_up': kwargs.get('last_seen_up', faketime),
        'created_at': kwargs.get('created_at', faketime),
        'updated_at': kwargs.get('updated_at', faketime),
        'availability_zone': kwargs.get('availability_zone', 'fake-zone'),
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import exceptions as cinder_exception
from os_brick import exception as os_brick_exception
from os_brick.initiator import connector as brick_connector
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils

from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun import objects
from zun.volume import cinder_api as cinder


LOG = logging.getLogger(__name__)

CONF = zun.conf.CONF


def get_volume_connector_properties():
    """Wrapper to automatically set root_helper in brick calls.

    :param multipath: A boolean indicating whether the connector can
                      support multipath.
    :param enforce_multipath: If True, it raises exception when multipath=True
                              is specified but multipathd is not running.
                              If False, it falls back to multipath=False
                              when multipathd is not running.
    """

    return brick_connector.get_connector_properties(
        None,
        CONF.my_block_storage_ip,
        CONF.volume.use_multipath,
        enforce_multipath=True,
        host=CONF.host)


def get_volume_connector(protocol, driver=None,
                         device_scan_attempts=3,
                         *args, **kwargs):
    """Wrapper to get a brick connector object.

    This automatically populates the required protocol as well
    as the root_helper needed to execute commands.
    """

    if protocol.upper() == "RBD":
        kwargs['do_local_attach'] = True
    return brick_connector.InitiatorConnector.factory(
        protocol, None,
        driver=driver,
        use_multipath=CONF.volume.use_multipath,
        device_scan_attempts=device_scan_attempts,
        *args, **kwargs)


class CinderWorkflow(object):

    def __init__(self, context):
        self.context = context
        self.cinder_api = cinder.CinderAPI(self.context)

    def attach_volume(self, volmap):
        try:
            return self._do_attach_volume(self.cinder_api, volmap)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Failed to attach volume %(volume_id)s",
                              {'volume_id': volmap.cinder_volume_id})
                self.cinder_api.unreserve_volume(volmap.cinder_volume_id)

    def _do_attach_volume(self, cinder_api, volmap):
        volume_id = volmap.cinder_volume_id
        container_uuid = volmap.container_uuid

        cinder_api.reserve_volume(volume_id)
        conn_info = cinder_api.initialize_connection(
            volume_id,
            get_volume_connector_properties())
        LOG.info("Get connection information %s", conn_info)

        try:
            device_info = self._connect_volume(conn_info)
            LOG.info("Get device_info after connect to "
                     "volume %s", device_info)
        except Exception:
            with excutils.save_and_reraise_exception():
                cinder_api.terminate_connection(
                    volume_id, get_volume_connector_properties())

        conn_info['data']['device_path'] = device_info['path']
        mountpoint = device_info['path']
        try:
            volmap.connection_info = jsonutils.dumps(conn_info)
        except TypeError:
            pass
        # NOTE(hongbin): save connection_info in the database
        # before calling cinder_api.attach because the volume status
        # will go to 'in-use' then caller immediately try to detach
        # the volume and connection_info is required for detach.
        volmap.save()

        try:
            cinder_api.attach(volume_id=volume_id,
                              mountpoint=mountpoint,
                              hostname=CONF.host,
                              container_uuid=container_uuid)
            LOG.info("Attach volume to this server successfully")
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    self._disconnect_volume(conn_info)
                except os_brick_exception.VolumeDeviceNotFound as exc:
                    LOG.warning('Ignoring VolumeDeviceNotFound: %s', exc)

                cinder_api.terminate_connection(
                    volume_id, get_volume_connector_properties())

                # Cinder-volume might have completed volume attach. So
                # we should detach the volume. If the attach did not
                # happen, the detach request will be ignored.
                cinder_api.detach(volmap)

        return device_info['path']

    def _connect_volume(self, conn_info):
        protocol = conn_info['driver_volume_type']
        connector = get_volume_connector(protocol)
        device_info = connector.connect_volume(conn_info['data'])
        return device_info

    def _disconnect_volume(self, conn_info):
        protocol = conn_info['driver_volume_type']
        connector = get_volume_connector(protocol)
        connector.disconnect_volume(conn_info['data'], None)

    def detach_volume(self, context, volmap):
        volume_id = volmap.cinder_volume_id
        try:
            self.cinder_api.begin_detaching(volume_id)
        except cinder_exception.BadRequest as e:
            raise exception.Invalid(_("Invalid volume: %s") %
                                    str(e))

        conn_info = jsonutils.loads(volmap.connection_info)
        if not self._volume_connection_keep(context, volume_id):
            try:
                self._disconnect_volume(conn_info)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception('Failed to disconnect volume %(volume_id)s',
                                  {'volume_id': volume_id})
                    self.cinder_api.roll_detaching(volume_id)

            self.cinder_api.terminate_connection(
                volume_id, get_volume_connector_properties())
        self.cinder_api.detach(volmap)

    def _volume_connection_keep(self, context, volume_id):
        host = CONF.host
        db_volumes = objects.VolumeMapping.list_by_cinder_volume(
            context, volume_id)
        volume_hosts = [db_volume.host for db_volume in db_volumes]

        if volume_hosts.count(host) == 1:
            return False
        return True

    def delete_volume(self, volmap):
        volume_id = volmap.cinder_volume_id
        try:
            self.cinder_api.delete_volume(volume_id)
        except cinder_exception as e:
            raise exception.Invalid(_("Delete Volume failed: %s") %
                                    str(e))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from zun.criapi import api_pb2 as zun_dot_criapi_dot_api__pb2


class RuntimeServiceStub(object):
    """Runtime service defines the public APIs for remote container runtimes
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Version = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/Version',
                request_serializer=zun_dot_criapi_dot_api__pb2.VersionRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.VersionResponse.FromString,
                )
        self.RunPodSandbox = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/RunPodSandbox',
                request_serializer=zun_dot_criapi_dot_api__pb2.RunPodSandboxRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.RunPodSandboxResponse.FromString,
                )
        self.StopPodSandbox = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/StopPodSandbox',
                request_serializer=zun_dot_criapi_dot_api__pb2.StopPodSandboxRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.StopPodSandboxResponse.FromString,
                )
        self.RemovePodSandbox = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/RemovePodSandbox',
                request_serializer=zun_dot_criapi_dot_api__pb2.RemovePodSandboxRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.RemovePodSandboxResponse.FromString,
                )
        self.PodSandboxStatus = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/PodSandboxStatus',
                request_serializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatusRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatusResponse.FromString,
                )
        self.ListPodSandbox = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ListPodSandbox',
                request_serializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxResponse.FromString,
                )
        self.CreateContainer = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/CreateContainer',
                request_serializer=zun_dot_criapi_dot_api__pb2.CreateContainerRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.CreateContainerResponse.FromString,
                )
        self.StartContainer = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/StartContainer',
                request_serializer=zun_dot_criapi_dot_api__pb2.StartContainerRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.StartContainerResponse.FromString,
                )
        self.StopContainer = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/StopContainer',
                request_serializer=zun_dot_criapi_dot_api__pb2.StopContainerRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.StopContainerResponse.FromString,
                )
        self.RemoveContainer = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/RemoveContainer',
                request_serializer=zun_dot_criapi_dot_api__pb2.RemoveContainerRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.RemoveContainerResponse.FromString,
                )
        self.ListContainers = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ListContainers',
                request_serializer=zun_dot_criapi_dot_api__pb2.ListContainersRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ListContainersResponse.FromString,
                )
        self.ContainerStatus = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ContainerStatus',
                request_serializer=zun_dot_criapi_dot_api__pb2.ContainerStatusRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ContainerStatusResponse.FromString,
                )
        self.UpdateContainerResources = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/UpdateContainerResources',
                request_serializer=zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesResponse.FromString,
                )
        self.ReopenContainerLog = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ReopenContainerLog',
                request_serializer=zun_dot_criapi_dot_api__pb2.ReopenContainerLogRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ReopenContainerLogResponse.FromString,
                )
        self.ExecSync = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ExecSync',
                request_serializer=zun_dot_criapi_dot_api__pb2.ExecSyncRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ExecSyncResponse.FromString,
                )
        self.Exec = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/Exec',
                request_serializer=zun_dot_criapi_dot_api__pb2.ExecRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ExecResponse.FromString,
                )
        self.Attach = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/Attach',
                request_serializer=zun_dot_criapi_dot_api__pb2.AttachRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.AttachResponse.FromString,
                )
        self.PortForward = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/PortForward',
                request_serializer=zun_dot_criapi_dot_api__pb2.PortForwardRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.PortForwardResponse.FromString,
                )
        self.ContainerStats = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ContainerStats',
                request_serializer=zun_dot_criapi_dot_api__pb2.ContainerStatsRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ContainerStatsResponse.FromString,
                )
        self.ListContainerStats = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ListContainerStats',
                request_serializer=zun_dot_criapi_dot_api__pb2.ListContainerStatsRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ListContainerStatsResponse.FromString,
                )
        self.PodSandboxStats = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/PodSandboxStats',
                request_serializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatsRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatsResponse.FromString,
                )
        self.ListPodSandboxStats = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/ListPodSandboxStats',
                request_serializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsResponse.FromString,
                )
        self.UpdateRuntimeConfig = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/UpdateRuntimeConfig',
                request_serializer=zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigResponse.FromString,
                )
        self.Status = channel.unary_unary(
                '/runtime.v1alpha2.RuntimeService/Status',
                request_serializer=zun_dot_criapi_dot_api__pb2.StatusRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.StatusResponse.FromString,
                )


class RuntimeServiceServicer(object):
    """Runtime service defines the public APIs for remote container runtimes
    """

    def Version(self, request, context):
        """Version returns the runtime name, runtime version, and runtime API version.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunPodSandbox(self, request, context):
        """RunPodSandbox creates and starts a pod-level sandbox. Runtimes must ensure
        the sandbox is in the ready state on success.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopPodSandbox(self, request, context):
        """StopPodSandbox stops any running process that is part of the sandbox and
        reclaims network resources (e.g., IP addresses) allocated to the sandbox.
        If there are any running containers in the sandbox, they must be forcibly
        terminated.
        This call is idempotent, and must not return an error if all relevant
        resources have already been reclaimed. kubelet will call StopPodSandbox
        at least once before calling RemovePodSandbox. It will also attempt to
        reclaim resources eagerly, as soon as a sandbox is not needed. Hence,
        multiple StopPodSandbox calls are expected.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemovePodSandbox(self, request, context):
        """RemovePodSandbox removes the sandbox. If there are any running containers
        in the sandbox, they must be forcibly terminated and removed.
        This call is idempotent, and must not return an error if the sandbox has
        already been removed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PodSandboxStatus(self, request, context):
        """PodSandboxStatus returns the status of the PodSandbox. If the PodSandbox is not
        present, returns an error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListPodSandbox(self, request, context):
        """ListPodSandbox returns a list of PodSandboxes.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateContainer(self, request, context):
        """CreateContainer creates a new container in specified PodSandbox
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartContainer(self, request, context):
        """StartContainer starts the container.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopContainer(self, request, context):
        """StopContainer stops a running container with a grace period (i.e., timeout).
        This call is idempotent, and must not return an error if the container has
        already been stopped.
        The runtime must forcibly kill the container after the grace period is
        reached.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveContainer(self, request, context):
        """RemoveContainer removes the container. If the container is running, the
        container must be forcibly removed.
        This call is idempotent, and must not return an error if the container has
        already been removed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListContainers(self, request, context):
        """ListContainers lists all containers by filters.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ContainerStatus(self, request, context):
        """ContainerStatus returns status of the container. If the container is not
        present, returns an error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateContainerResources(self, request, context):
        """UpdateContainerResources updates ContainerConfig of the container.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReopenContainerLog(self, request, context):
        """ReopenContainerLog asks runtime to reopen the stdout/stderr log file
        for the container. This is often called after the log file has been
        rotated. If the container is not running, container runtime can choose
        to either create a new log file and return nil, or return an error.
        Once it returns error, new container log file MUST NOT be created.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecSync(self, request, context):
        """ExecSync runs a command in a container synchronously.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Exec(self, request, context):
        """Exec prepares a streaming endpoint to execute a command in the container.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Attach(self, request, context):
        """Attach prepares a streaming endpoint to attach to a running container.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PortForward(self, request, context):
        """PortForward prepares a streaming endpoint to forward ports from a PodSandbox.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ContainerStats(self, request, context):
        """ContainerStats returns stats of the container. If the container does not
        exist, the call returns an error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListContainerStats(self, request, context):
        """ListContainerStats returns stats of all running containers.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PodSandboxStats(self, request, context):
        """PodSandboxStats returns stats of the pod sandbox. If the pod sandbox does not
        exist, the call returns an error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListPodSandboxStats(self, request, context):
        """ListPodSandboxStats returns stats of the pod sandboxes matching a filter.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateRuntimeConfig(self, request, context):
        """UpdateRuntimeConfig updates the runtime configuration based on the given request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Status(self, request, context):
        """Status returns the status of the runtime.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RuntimeServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.VersionRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.VersionResponse.SerializeToString,
            ),
            'RunPodSandbox': grpc.unary_unary_rpc_method_handler(
                    servicer.RunPodSandbox,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.RunPodSandboxRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.RunPodSandboxResponse.SerializeToString,
            ),
            'StopPodSandbox': grpc.unary_unary_rpc_method_handler(
                    servicer.StopPodSandbox,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.StopPodSandboxRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.StopPodSandboxResponse.SerializeToString,
            ),
            'RemovePodSandbox': grpc.unary_unary_rpc_method_handler(
                    servicer.RemovePodSandbox,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.RemovePodSandboxRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.RemovePodSandboxResponse.SerializeToString,
            ),
            'PodSandboxStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.PodSandboxStatus,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatusRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatusResponse.SerializeToString,
            ),
            'ListPodSandbox': grpc.unary_unary_rpc_method_handler(
                    servicer.ListPodSandbox,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxResponse.SerializeToString,
            ),
            'CreateContainer': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateContainer,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.CreateContainerRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.CreateContainerResponse.SerializeToString,
            ),
            'StartContainer': grpc.unary_unary_rpc_method_handler(
                    servicer.StartContainer,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.StartContainerRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.StartContainerResponse.SerializeToString,
            ),
            'StopContainer': grpc.unary_unary_rpc_method_handler(
                    servicer.StopContainer,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.StopContainerRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.StopContainerResponse.SerializeToString,
            ),
            'RemoveContainer': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveContainer,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.RemoveContainerRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.RemoveContainerResponse.SerializeToString,
            ),
            'ListContainers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListContainers,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ListContainersRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ListContainersResponse.SerializeToString,
            ),
            'ContainerStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.ContainerStatus,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ContainerStatusRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ContainerStatusResponse.SerializeToString,
            ),
            'UpdateContainerResources': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateContainerResources,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesResponse.SerializeToString,
            ),
            'ReopenContainerLog': grpc.unary_unary_rpc_method_handler(
                    servicer.ReopenContainerLog,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ReopenContainerLogRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ReopenContainerLogResponse.SerializeToString,
            ),
            'ExecSync': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecSync,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ExecSyncRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ExecSyncResponse.SerializeToString,
            ),
            'Exec': grpc.unary_unary_rpc_method_handler(
                    servicer.Exec,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ExecRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ExecResponse.SerializeToString,
            ),
            'Attach': grpc.unary_unary_rpc_method_handler(
                    servicer.Attach,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.AttachRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.AttachResponse.SerializeToString,
            ),
            'PortForward': grpc.unary_unary_rpc_method_handler(
                    servicer.PortForward,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.PortForwardRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.PortForwardResponse.SerializeToString,
            ),
            'ContainerStats': grpc.unary_unary_rpc_method_handler(
                    servicer.ContainerStats,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ContainerStatsRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ContainerStatsResponse.SerializeToString,
            ),
            'ListContainerStats': grpc.unary_unary_rpc_method_handler(
                    servicer.ListContainerStats,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ListContainerStatsRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ListContainerStatsResponse.SerializeToString,
            ),
            'PodSandboxStats': grpc.unary_unary_rpc_method_handler(
                    servicer.PodSandboxStats,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatsRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.PodSandboxStatsResponse.SerializeToString,
            ),
            'ListPodSandboxStats': grpc.unary_unary_rpc_method_handler(
                    servicer.ListPodSandboxStats,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsResponse.SerializeToString,
            ),
            'UpdateRuntimeConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateRuntimeConfig,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigResponse.SerializeToString,
            ),
            'Status': grpc.unary_unary_rpc_method_handler(
                    servicer.Status,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.StatusRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.StatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'runtime.v1alpha2.RuntimeService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RuntimeService(object):
    """Runtime service defines the public APIs for remote container runtimes
    """

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/Version',
            zun_dot_criapi_dot_api__pb2.VersionRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.VersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunPodSandbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/RunPodSandbox',
            zun_dot_criapi_dot_api__pb2.RunPodSandboxRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.RunPodSandboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StopPodSandbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/StopPodSandbox',
            zun_dot_criapi_dot_api__pb2.StopPodSandboxRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.StopPodSandboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemovePodSandbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/RemovePodSandbox',
            zun_dot_criapi_dot_api__pb2.RemovePodSandboxRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.RemovePodSandboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PodSandboxStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/PodSandboxStatus',
            zun_dot_criapi_dot_api__pb2.PodSandboxStatusRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.PodSandboxStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListPodSandbox(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ListPodSandbox',
            zun_dot_criapi_dot_api__pb2.ListPodSandboxRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ListPodSandboxResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateContainer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/CreateContainer',
            zun_dot_criapi_dot_api__pb2.CreateContainerRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.CreateContainerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StartContainer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/StartContainer',
            zun_dot_criapi_dot_api__pb2.StartContainerRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.StartContainerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StopContainer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/StopContainer',
            zun_dot_criapi_dot_api__pb2.StopContainerRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.StopContainerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveContainer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/RemoveContainer',
            zun_dot_criapi_dot_api__pb2.RemoveContainerRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.RemoveContainerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListContainers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ListContainers',
            zun_dot_criapi_dot_api__pb2.ListContainersRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ListContainersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ContainerStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ContainerStatus',
            zun_dot_criapi_dot_api__pb2.ContainerStatusRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ContainerStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateContainerResources(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/UpdateContainerResources',
            zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.UpdateContainerResourcesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReopenContainerLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ReopenContainerLog',
            zun_dot_criapi_dot_api__pb2.ReopenContainerLogRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ReopenContainerLogResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExecSync(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ExecSync',
            zun_dot_criapi_dot_api__pb2.ExecSyncRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ExecSyncResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Exec(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/Exec',
            zun_dot_criapi_dot_api__pb2.ExecRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ExecResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Attach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/Attach',
            zun_dot_criapi_dot_api__pb2.AttachRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.AttachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PortForward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/PortForward',
            zun_dot_criapi_dot_api__pb2.PortForwardRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.PortForwardResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ContainerStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ContainerStats',
            zun_dot_criapi_dot_api__pb2.ContainerStatsRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ContainerStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListContainerStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ListContainerStats',
            zun_dot_criapi_dot_api__pb2.ListContainerStatsRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ListContainerStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PodSandboxStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/PodSandboxStats',
            zun_dot_criapi_dot_api__pb2.PodSandboxStatsRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.PodSandboxStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListPodSandboxStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/ListPodSandboxStats',
            zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ListPodSandboxStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateRuntimeConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/UpdateRuntimeConfig',
            zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.UpdateRuntimeConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Status(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.RuntimeService/Status',
            zun_dot_criapi_dot_api__pb2.StatusRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ImageServiceStub(object):
    """ImageService defines the public APIs for managing images.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListImages = channel.unary_unary(
                '/runtime.v1alpha2.ImageService/ListImages',
                request_serializer=zun_dot_criapi_dot_api__pb2.ListImagesRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ListImagesResponse.FromString,
                )
        self.ImageStatus = channel.unary_unary(
                '/runtime.v1alpha2.ImageService/ImageStatus',
                request_serializer=zun_dot_criapi_dot_api__pb2.ImageStatusRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ImageStatusResponse.FromString,
                )
        self.PullImage = channel.unary_unary(
                '/runtime.v1alpha2.ImageService/PullImage',
                request_serializer=zun_dot_criapi_dot_api__pb2.PullImageRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.PullImageResponse.FromString,
                )
        self.RemoveImage = channel.unary_unary(
                '/runtime.v1alpha2.ImageService/RemoveImage',
                request_serializer=zun_dot_criapi_dot_api__pb2.RemoveImageRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.RemoveImageResponse.FromString,
                )
        self.ImageFsInfo = channel.unary_unary(
                '/runtime.v1alpha2.ImageService/ImageFsInfo',
                request_serializer=zun_dot_criapi_dot_api__pb2.ImageFsInfoRequest.SerializeToString,
                response_deserializer=zun_dot_criapi_dot_api__pb2.ImageFsInfoResponse.FromString,
                )


class ImageServiceServicer(object):
    """ImageService defines the public APIs for managing images.
    """

    def ListImages(self, request, context):
        """ListImages lists existing images.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImageStatus(self, request, context):
        """ImageStatus returns the status of the image. If the image is not
        present, returns a response with ImageStatusResponse.Image set to
        nil.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PullImage(self, request, context):
        """PullImage pulls an image with authentication config.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveImage(self, request, context):
        """RemoveImage removes the image.
        This call is idempotent, and must not return an error if the image has
        already been removed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImageFsInfo(self, request, context):
        """ImageFSInfo returns information of the filesystem that is used to store images.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ImageServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListImages': grpc.unary_unary_rpc_method_handler(
                    servicer.ListImages,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ListImagesRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ListImagesResponse.SerializeToString,
            ),
            'ImageStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.ImageStatus,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ImageStatusRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ImageStatusResponse.SerializeToString,
            ),
            'PullImage': grpc.unary_unary_rpc_method_handler(
                    servicer.PullImage,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.PullImageRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.PullImageResponse.SerializeToString,
            ),
            'RemoveImage': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveImage,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.RemoveImageRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.RemoveImageResponse.SerializeToString,
            ),
            'ImageFsInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.ImageFsInfo,
                    request_deserializer=zun_dot_criapi_dot_api__pb2.ImageFsInfoRequest.FromString,
                    response_serializer=zun_dot_criapi_dot_api__pb2.ImageFsInfoResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'runtime.v1alpha2.ImageService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ImageService(object):
    """ImageService defines the public APIs for managing images.
    """

    @staticmethod
    def ListImages(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.ImageService/ListImages',
            zun_dot_criapi_dot_api__pb2.ListImagesRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ListImagesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImageStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.ImageService/ImageStatus',
            zun_dot_criapi_dot_api__pb2.ImageStatusRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ImageStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PullImage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.ImageService/PullImage',
            zun_dot_criapi_dot_api__pb2.PullImageRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.PullImageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveImage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.ImageService/RemoveImage',
            zun_dot_criapi_dot_api__pb2.RemoveImageRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.RemoveImageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImageFsInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/runtime.v1alpha2.ImageService/ImageFsInfo',
            zun_dot_criapi_dot_api__pb2.ImageFsInfoRequest.SerializeToString,
            zun_dot_criapi_dot_api__pb2.ImageFsInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import docker
from oslo_log import log as logging

import zun.conf
from zun.tests import base
from zun.tests.fullstack import utils


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


class BaseFullStackTestCase(base.TestCase):

    def setUp(self):
        super(BaseFullStackTestCase, self).setUp()

        self.docker = docker.APIClient(base_url='tcp://0.0.0.0:2375')
        try:
            self.zun = utils.get_zun_client_from_env()
        except Exception as e:
            # We may missing or didn't source configured openrc file.
            message = ("Missing environment variable %s in your local."
                       "Please add it and also check other missing "
                       "environment variables. After that please source "
                       "the openrc file. "
                       "Trying credentials from DevStack cloud.yaml ...")
            LOG.warning(message, e.args[0])
            self.zun = utils.get_zun_client_from_creds()

    def ensure_container_deleted(self, container_id):
        def is_container_deleted():
            containers = self.zun.containers.list()
            container_ids = [c.uuid for c in containers]
            if container_id in container_ids:
                return False
            else:
                return True
        utils.wait_for_condition(is_container_deleted)

    def ensure_container_in_desired_state(self, container_id, status):
        def is_container_in_desired_state():
            c = self.zun.containers.get(container_id)
            if c.status == status:
                return True
            else:
                return False
        utils.wait_for_condition(is_container_in_desired_state, timeout=300)

    def _get_container_in_docker(self, container):
        return self.docker.inspect_container('zun-' + container.uuid)

    def _get_state_in_docker(self, container):
        container = self.docker.inspect_container('zun-' + container.uuid)
        status = container.get('State')
        if status.get('Error') is True:
            return 'Error'
        elif status.get('Paused'):
            return 'Paused'
        elif status.get('Running'):
            return 'Running'
        elif status.get('Status') == 'created':
            return 'Created'
        else:
            return 'Stopped'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from testtools.matchers import HasLength

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestComputeNodeObject(base.DbTestCase):

    def setUp(self):
        super(TestComputeNodeObject, self).setUp()
        self.fake_numa_topology = utils.get_test_numa_topology()
        self.fake_compute_node = utils.get_test_compute_node(
            numa_topology=self.fake_numa_topology)

    def test_get_by_uuid(self):
        uuid = self.fake_compute_node['uuid']
        with mock.patch.object(self.dbapi, 'get_compute_node',
                               autospec=True) as mock_get_compute_node:
            mock_get_compute_node.return_value = self.fake_compute_node
            compute_node = objects.ComputeNode.get_by_uuid(self.context, uuid)
            mock_get_compute_node.assert_called_once_with(
                self.context, uuid)
            self.assertEqual(self.context, compute_node._context)

    def test_get_by_name(self):
        hostname = self.fake_compute_node['hostname']
        with mock.patch.object(self.dbapi, 'get_compute_node_by_hostname',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_compute_node
            compute_node = objects.ComputeNode.get_by_name(
                self.context, hostname)
            mock_get.assert_called_once_with(self.context, hostname)
            self.assertEqual(self.context, compute_node._context)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'list_compute_nodes',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_compute_node]
            compute_nodes = objects.ComputeNode.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(compute_nodes, HasLength(1))
            self.assertIsInstance(compute_nodes[0], objects.ComputeNode)
            self.assertEqual(self.context, compute_nodes[0]._context)

    def test_list_with_filters(self):
        with mock.patch.object(self.dbapi, 'list_compute_nodes',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_compute_node]
            filt = {'hostname': 'test'}
            compute_nodes = objects.ComputeNode.list(
                self.context, filters=filt)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(compute_nodes, HasLength(1))
            self.assertIsInstance(compute_nodes[0], objects.ComputeNode)
            self.assertEqual(self.context, compute_nodes[0]._context)
            mock_get_list.assert_called_once_with(
                self.context, filters=filt, limit=None, marker=None,
                sort_key=None, sort_dir=None)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_compute_node',
                               autospec=True) as mock_create:
            mock_create.return_value = self.fake_compute_node
            compute_node_dict = dict(self.fake_compute_node)
            compute_node_dict['numa_topology'] = objects.NUMATopology\
                ._from_dict(compute_node_dict['numa_topology'])
            compute_node = objects.ComputeNode(
                self.context, **compute_node_dict)
            compute_node.create(self.context)
            mock_create.assert_called_once_with(
                self.context, self.fake_compute_node)
            self.assertEqual(self.context, compute_node._context)

    def test_destroy(self):
        uuid = self.fake_compute_node['uuid']
        with mock.patch.object(self.dbapi, 'get_compute_node',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_compute_node
            with mock.patch.object(self.dbapi, 'destroy_compute_node',
                                   autospec=True) as mock_destroy:
                compute_node = objects.ComputeNode.get_by_uuid(
                    self.context, uuid)
                compute_node.destroy()
                mock_get.assert_called_once_with(self.context, uuid)
                mock_destroy.assert_called_once_with(None, uuid)
                self.assertEqual(self.context, compute_node._context)

    def test_save(self):
        uuid = self.fake_compute_node['uuid']
        with mock.patch.object(self.dbapi, 'get_compute_node',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_compute_node
            with mock.patch.object(self.dbapi, 'update_compute_node',
                                   autospec=True) as mock_update:
                compute_node = objects.ComputeNode.get_by_uuid(
                    self.context, uuid)
                compute_node.hostname = 'myhostname'
                compute_node.save()

                mock_get.assert_called_once_with(self.context, uuid)
                mock_update.assert_called_once_with(
                    None, uuid,
                    {'hostname': 'myhostname'})
                self.assertEqual(self.context, compute_node._context)

    def test_refresh(self):
        uuid = self.fake_compute_node['uuid']
        hostname = self.fake_compute_node['hostname']
        new_hostname = 'myhostname'
        returns = [dict(self.fake_compute_node, hostname=hostname),
                   dict(self.fake_compute_node, hostname=new_hostname)]
        expected = [mock.call(self.context, uuid),
                    mock.call(self.context, uuid)]
        with mock.patch.object(self.dbapi, 'get_compute_node',
                               side_effect=returns,
                               autospec=True) as mock_get:
            compute_node = objects.ComputeNode.get_by_uuid(self.context, uuid)
            self.assertEqual(hostname, compute_node.hostname)
            compute_node.refresh()
            self.assertEqual(new_hostname, compute_node.hostname)
            self.assertEqual(expected, mock_get.call_args_list)
            self.assertEqual(self.context, compute_node._context)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1.access import access as ka_access
from keystoneauth1.identity import access as ka_access_plugin
from keystoneauth1.identity import v3 as ka_v3
from keystoneauth1 import loading as ka_loading
from keystoneclient.v3 import client as kc_v3
from oslo_log import log as logging

from zun.common import exception
import zun.conf
from zun.conf import keystone as ksconf


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


class KeystoneClientV3(object):
    """Keystone client wrapper so we can encapsulate logic in one place."""

    def __init__(self, context):
        self.context = context
        self._client = None
        self._session = None
        self._auth = None

    @property
    def auth_url(self):
        # FIXME(pauloewerton): auth_url should be retrieved from keystone_auth
        # section by default
        url = CONF[ksconf.CFG_LEGACY_GROUP].www_authenticate_uri or \
            CONF[ksconf.CFG_LEGACY_GROUP].auth_uri
        return url.replace('v2.0', 'v3')

    @property
    def auth_token(self):
        return self.session.get_token()

    @property
    def session(self):
        if self._session:
            return self._session
        auth = self._get_auth()
        session = self._get_session(auth)
        self._session = session
        return session

    @property
    def auth(self):
        if self._auth:
            return self._auth
        auth = self._get_auth()
        self._auth = auth
        return auth

    def _get_session(self, auth):
        session = ka_loading.load_session_from_conf_options(
            CONF, ksconf.CFG_GROUP, auth=auth)
        return session

    def _get_auth(self):
        if self.context.auth_token_info:
            access_info = ka_access.create(body=self.context.auth_token_info,
                                           auth_token=self.context.auth_token)
            auth = ka_access_plugin.AccessInfoPlugin(access_info)
        elif self.context.auth_token:
            auth = ka_v3.Token(auth_url=self.auth_url,
                               token=self.context.auth_token)
        elif self.context.is_admin:
            auth = ka_loading.load_auth_from_conf_options(CONF,
                                                          ksconf.CFG_GROUP)
        else:
            msg = ('Keystone API connection failed: no password, '
                   'trust_id or token found.')
            LOG.error(msg)
            raise exception.AuthorizationFailure(client='keystone',
                                                 message='reason %s' % msg)

        return auth

    @property
    def client(self):
        if self._client:
            return self._client
        client = kc_v3.Client(session=self.session)
        self._client = client
        return client



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add cni_metadata to container

Revision ID: 47d79ffdc582
Revises: c2052ead4f95
Create Date: 2019-12-30 17:40:10.199545

"""

# revision identifiers, used by Alembic.
revision = '47d79ffdc582'
down_revision = 'c2052ead4f95'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def MediumText():
    return sa.Text().with_variant(sa.dialects.mysql.MEDIUMTEXT(), 'mysql')


def upgrade():
    op.add_column('container',
                  sa.Column('cni_metadata', MediumText(), nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

common_quota = {
    'type': ['integer', 'string'],
    'pattern': '^-?[0-9]+$',
    # -1 is a flag value for unlimited
    'minimum': -1
}

quota_resources = {
    'containers': common_quota,
    'memory': common_quota,
    'cpu': common_quota,
    'disk': common_quota
}

query_param_update = {
    'type': 'object',
    'properties': quota_resources,
    'additionalProperties': False
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import time

from neutron_lib import constants as n_const
from neutronclient.common import exceptions as n_exceptions
from neutronclient.neutron import v2_0 as neutronv20
from oslo_log import log as logging
from oslo_utils import uuidutils

from zun.common import clients
from zun.common import consts
from zun.common import context as zun_context
from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun.objects import fields as obj_fields
from zun.pci import manager as pci_manager
from zun.pci import utils as pci_utils
from zun.pci import whitelist as pci_whitelist


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


class NeutronAPI(object):

    def __init__(self, context):
        self.context = context
        self.client = clients.OpenStackClients(self.context).neutron()
        self.admin_client = None
        self.pci_whitelist = pci_whitelist.Whitelist(
            CONF.pci.passthrough_whitelist)
        self.last_neutron_extension_sync = None
        self.extensions = {}

    def __getattr__(self, key):
        return getattr(self.client, key)

    def _get_admin_client(self):
        if self.admin_client is None:
            context = zun_context.get_admin_context()
            self.admin_client = clients.OpenStackClients(context).neutron()
        return self.admin_client

    def update_port(self, port, body=None, admin=False):
        if admin:
            client = self._get_admin_client()
        else:
            client = self.client
        return client.update_port(port, body)

    def create_port(self, body=None, admin=False):
        if admin:
            client = self._get_admin_client()
        else:
            client = self.client
        return client.create_port(body)

    def create_or_update_port(self, container, network_uuid,
                              requested_network, device_owner,
                              security_groups=None, set_binding_host=False):
        if requested_network.get('port'):
            neutron_port_id = requested_network.get('port')
            # update device_id in port
            port_req_body = {'port': {'device_id': container.uuid}}
            if set_binding_host:
                port_req_body['port']['device_owner'] = device_owner
                port_req_body['port'][consts.BINDING_HOST_ID] = container.host
            neutron_port = self.update_port(neutron_port_id, port_req_body,
                                            admin=True)

            # If there is pci_request_id, it should be a sriov port.
            # populate pci related info.
            pci_request_id = requested_network.get('pci_request_id')
            if pci_request_id:
                self._populate_neutron_extension_values(container,
                                                        pci_request_id,
                                                        port_req_body)
                self._populate_pci_mac_address(container,
                                               pci_request_id,
                                               port_req_body)
                # NOTE(hongbin): Use admin context here because non-admin
                # context might not be able to update some attributes
                # (i.e. binding:profile).
                neutron_port = self.update_port(neutron_port_id, port_req_body,
                                                admin=True)
        else:
            port_dict = {
                'network_id': network_uuid,
                'tenant_id': self.context.project_id,
                'device_id': container.uuid,
            }
            if set_binding_host:
                port_dict['device_owner'] = device_owner
                port_dict[consts.BINDING_HOST_ID] = container.host
            ip_addr = requested_network.get("fixed_ip")
            if ip_addr:
                port_dict['fixed_ips'] = [{'ip_address': ip_addr}]
            if security_groups is not None:
                port_dict['security_groups'] = security_groups
            neutron_port = self.create_port({'port': port_dict}, admin=True)

        neutron_port = neutron_port['port']
        preserve_on_delete = requested_network['preserve_on_delete']
        addresses = []
        for fixed_ip in neutron_port['fixed_ips']:
            ip_address = fixed_ip['ip_address']
            ip = ipaddress.ip_address(str(ip_address))
            if ip.version == 4:
                addresses.append({
                    'addr': ip_address,
                    'version': 4,
                    'port': neutron_port['id'],
                    'subnet_id': fixed_ip['subnet_id'],
                    'preserve_on_delete': preserve_on_delete
                })
            else:
                addresses.append({
                    'addr': ip_address,
                    'version': 6,
                    'port': neutron_port['id'],
                    'subnet_id': fixed_ip['subnet_id'],
                    'preserve_on_delete': preserve_on_delete
                })

        return addresses, neutron_port

    def delete_or_unbind_ports(self, ports, ports_to_delete):
        for port_id in ports:
            try:
                if port_id in ports_to_delete:
                    self.delete_port(port_id)
                else:
                    self._unbind_port(port_id)
            except n_exceptions.PortNotFoundClient:
                LOG.warning('Maybe your libnetwork distribution do not '
                            'have patch https://review.openstack.org/#/c/'
                            '441024/ or neutron tag extension does not '
                            'supported or not enabled.')

    def _unbind_port(self, port_id):
        port_req_body = {'port': {'device_id': '', 'device_owner': ''}}
        port_req_body['port'][consts.BINDING_HOST_ID] = None
        try:
            port = self.get_neutron_port(port_id)
        except exception.PortNotFound:
            LOG.debug('Unable to show port %s as it no longer '
                      'exists.', port_id)
            return
        except Exception:
            # NOTE: In case we can't retrieve the binding:profile assume
            # that they are empty
            LOG.exception("Unable to get binding:profile for port '%s'",
                          port_id)
            port_profile = {}
        else:
            port_profile = port.get(consts.BINDING_PROFILE, {})
        # NOTE: We're doing this to remove the binding information
        # for the physical device but don't want to overwrite the other
        # information in the binding profile.
        for profile_key in ('pci_vendor_info', 'pci_slot'):
            if profile_key in port_profile:
                del port_profile[profile_key]
        port_req_body['port'][consts.BINDING_PROFILE] = port_profile

        try:
            # Requires admin creds to set port bindings
            self.update_port(port_id, port_req_body, admin=True)
        except exception.PortNotFound:
            LOG.debug('Unable to unbind port %s as it no longer '
                      'exists.', port_id)
        except Exception:
            LOG.exception("Unable to clear device ID for port '%s'",
                          port_id)

    def _refresh_neutron_extensions_cache(self):
        """Refresh the neutron extensions cache when necessary."""
        if (not self.last_neutron_extension_sync or
            ((time.time() - self.last_neutron_extension_sync) >=
             CONF.neutron.extension_sync_interval)):
            extensions_list = self.neutron_api.list_extensions()['extensions']
            self.last_neutron_extension_sync = time.time()
            self.extensions.clear()
            self.extensions = {ext['name']: ext for ext in extensions_list}

    def _populate_neutron_extension_values(self, container, pci_request_id,
                                           port_req_body):
        self._refresh_neutron_extensions_cache()
        if "Port Binding" in self.extensions:
            self._populate_neutron_binding_profile(container, pci_request_id,
                                                   port_req_body)

    def _populate_neutron_binding_profile(self, container, pci_request_id,
                                          port_req_body):
        """Populate neutron binding:profile.

        Populate it with SR-IOV related information
        """
        if pci_request_id:
            pci_dev = pci_manager.get_container_pci_devs(
                container, pci_request_id).pop()
            profile = self._get_pci_device_profile(pci_dev)
            port_req_body['port'][consts.BINDING_PROFILE] = profile

    def _get_pci_device_profile(self, pci_dev):
        dev_spec = self.pci_whitelist.get_devspec(pci_dev)
        if dev_spec:
            return {'pci_vendor_info': "%s:%s" % (pci_dev.vendor_id,
                                                  pci_dev.product_id),
                    'pci_slot': pci_dev.address,
                    'physical_network':
                        dev_spec.get_tags().get('physical_network')}
        raise exception.PciDeviceNotFound(node_id=pci_dev.compute_node_uuid,
                                          address=pci_dev.address)

    def _populate_pci_mac_address(self, container, pci_request_id,
                                  port_req_body):
        """Add the updated MAC address value to the update_port request body.

        Currently this is done only for PF passthrough.
        """
        if pci_request_id is not None:
            pci_devs = pci_manager.get_container_pci_devs(
                container, pci_request_id)
            if len(pci_devs) != 1:
                # NOTE(ndipanov): We shouldn't ever get here since
                # InstancePCIRequest instances built from network requests
                # only ever index a single device, which needs to be
                # successfully claimed for this to be called as part of
                # allocate_networks method
                LOG.error("PCI request %(pci_request_id)s does not have a "
                          "unique device associated with it. Unable to "
                          "determine MAC address",
                          {'pci_request_id': pci_request_id},
                          container=container)
                return
            pci_dev = pci_devs[0]
            if pci_dev.dev_type == obj_fields.PciDeviceType.SRIOV_PF:
                try:
                    mac = pci_utils.get_mac_by_pci_address(pci_dev.address)
                except exception.PciDeviceNotFoundById as e:
                    LOG.error("Could not determine MAC address for %(addr)s, "
                              "error: %(e)s",
                              {"addr": pci_dev.address, "e": e},
                              container=container)
                else:
                    port_req_body['port']['mac_address'] = mac

    def find_resourceid_by_name_or_id(self, resource, name_or_id,
                                      project_id=None):
        return neutronv20.find_resourceid_by_name_or_id(
            self.client, resource, name_or_id, project_id)

    def get_available_network(self):
        search_opts = {'tenant_id': self.context.project_id, 'shared': False}
        # NOTE(kiennt): Pick shared network if no tenant network
        nets = self.list_networks(**search_opts).get('networks', []) or \
            self.list_networks(**{'shared': True}).get('networks', [])
        if not nets:
            raise exception.Conflict(_(
                "There is no neutron network available"))
        nets.sort(key=lambda x: x['created_at'])
        return nets[0]

    def get_neutron_network(self, network):
        if uuidutils.is_uuid_like(network):
            networks = self.list_networks(id=network)['networks']
        else:
            networks = self.list_networks(name=network)['networks']

        if len(networks) == 0:
            raise exception.NetworkNotFound(network=network)
        elif len(networks) > 1:
            raise exception.Conflict(_(
                'Multiple neutron networks exist with same name. '
                'Please use the uuid instead.'))

        network = networks[0]
        return network

    def get_neutron_port(self, port):
        if uuidutils.is_uuid_like(port):
            ports = self.list_ports(id=port)['ports']
        else:
            ports = self.list_ports(name=port)['ports']

        if len(ports) == 0:
            raise exception.PortNotFound(port=port)
        elif len(ports) > 1:
            raise exception.Conflict(_(
                'Multiple neutron ports exist with same name. '
                'Please use the uuid instead.'))

        port = ports[0]
        return port

    def ensure_neutron_port_usable(self, port, project_id=None):
        if project_id is None:
            project_id = self.context.project_id

        # Make sure the container has access to the port.
        if port['tenant_id'] != project_id:
            raise exception.PortNotUsable(port=port['id'])

        # Make sure the port isn't already attached to another
        # container or Nova instance.
        if port.get('device_id'):
            raise exception.PortInUse(port=port['id'])

        if port.get('status') in (n_const.PORT_STATUS_ACTIVE,
                                  n_const.PORT_STATUS_BUILD,
                                  n_const.PORT_STATUS_ERROR):
            raise exception.PortNotUsable(port=port['id'])

        # Make sure the port is usable
        binding_vif_type = port.get('binding:vif_type')
        if binding_vif_type == 'binding_failed':
            raise exception.PortBindingFailed(port=port['id'])

    def expose_ports(self, secgroup_id, ports):
        for port in ports:
            port, proto = port.split('/')
            secgroup_rule = {
                'security_group_id': secgroup_id,
                'direction': 'ingress',
                'port_range_min': port,
                'port_range_max': port,
                'protocol': proto,
                'remote_ip_prefix': '0.0.0.0/0',
            }

            try:
                self.create_security_group_rule({
                    'security_group_rule': secgroup_rule})
            except n_exceptions.NeutronClientException as ex:
                LOG.error("Error happened during creating a "
                          "Neutron security group "
                          "rule: %s", ex)
                self.delete_security_group(secgroup_id)
                raise exception.ZunException(_(
                    "Could not create required security group rules %s "
                    "for setting up exported port.") % secgroup_rule)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from random import Random


class NameGenerator(object):
    letters = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta',
               'eta', 'theta', 'iota', 'kappa', 'lambda', 'mu', 'nu',
               'xi', 'omicron', 'pi', 'rho', 'sigma', 'tau', 'upsilon',
               'phi', 'chi', 'psi', 'omega']

    def __init__(self):
        self.random = Random()

    def generate(self):
        """Generate a random name compose of a Greek leter and

        a number, like: beta_2.
        """

        letter = self.random.choice(self.letters)
        number = self.random.randint(1, 24)

        return letter + '-' + str(number)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add container_id column to container

Revision ID: 5971a6844738
Revises: 9fe371393a24
Create Date: 2016-08-05 17:38:05.231740

"""

# revision identifiers, used by Alembic.
revision = '5971a6844738'
down_revision = '9fe371393a24'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('container',
                  sa.Column('container_id', sa.String(length=255),
                            nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from zun.tests.unit.api import base as api_base


class TestQuotaClassController(api_base.FunctionalTest):
    def setUp(self):
        super(TestQuotaClassController, self).setUp()
        self.default_quotas = {
            'containers': 40,
            'cpu': 20,
            'memory': 51200,
            'disk': 100
        }

    @mock.patch('zun.common.policy.enforce', return_value=True)
    def test_put_quota(self, mock_policy):
        update_quota_dicts = {
            'containers': '50',
            'memory': '61440'
        }

        admin_project_id = 'fakeadminprojectid'
        path = '/quota_classes/' + admin_project_id
        response = self.put_json(path, update_quota_dicts)
        self.assertEqual(200, response.status_int)
        self.assertEqual(50, response.json['containers'])
        self.assertEqual(61440, response.json['memory'])

    @mock.patch('zun.common.policy.enforce', return_value=True)
    def test_get_quota(self, mock_policy):
        admin_project_id = 'fakeadminprojectid'
        path = '/quota_classes/' + admin_project_id
        response = self.get(path)
        self.assertEqual(200, response.status_int)
        self.assertEqual(self.default_quotas, response.json)
